__all__ = ['model_from_owl_str', 'model_from_owl_file', 'model_to_owl_str',
           'model_to_owl_file', 'model_from_owl_url', 'model_from_pc_query',
           'model_from_reactome', 'model_from_ecocyc', 'model_from_metacyc',
           'model_from_biocyc', 'model_from_humancyc', 'model_from_netpath',
           'model_from_owl_gz',
           ]

import gzip
import os
import pathlib

import requests
from lxml import etree
from typing import Any, Mapping, Optional, Union
from .biopax.model import BioPaxModel
from .xml_util import xml_to_str, xml_to_file
from .pc_client import graph_query


def model_from_owl_str(owl_str: str) -> BioPaxModel:
    """Return a BioPAX Model from an OWL string.

    Parameters
    ----------
    owl_str :
        A OWL string of BioPAX content.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model deserialized from the OWL string.
    """
    return BioPaxModel.from_xml(etree.fromstring(owl_str.encode('utf-8')))


def model_from_owl_file(fname: Union[str, pathlib.Path, os.PathLike],
                        encoding: Optional[str] = None) \
        -> BioPaxModel:
    """Return a BioPAX Model from an OWL string.

    Parameters
    ----------
    fname :
        A path to an OWL file of BioPAX content.
    encoding :
        The encoding type to be passed to :func:`open`.

    Returns
    -------
    :
        A BioPAX Model deserialized from the OWL file.
    """
    with open(fname, 'r', encoding=encoding) as fh:
        owl_str = fh.read()
        return model_from_owl_str(owl_str)


def model_from_owl_gz(path: Union[str, pathlib.Path, os.PathLike]) \
        -> BioPaxModel:
    """Return a BioPAX Model from an OWL file (gzipped).

    Parameters
    ----------
    path :
        A path to a gzipped OWL file of BioPAX content.

    Returns
    -------
    :
        A BioPAX Model deserialized from the OWL file.
    """
    with gzip.open(path, 'rt') as fh:
        return BioPaxModel.from_xml(etree.parse(fh).getroot())


def model_from_owl_url(url: str,
                       request_params: Optional[Mapping[str, Any]] = None) \
        -> BioPaxModel:
    """Return a BioPAX Model from an URL pointing to an OWL file.

    Parameters
    ----------
    url :
        A OWL URL with BioPAX content.
    request_params :
        Additional keyword arguments to pass to :func:`requests.get`

    Returns
    -------
    :
        A BioPAX Model deserialized from the OWL file.
    """
    request_params = {} if not request_params else request_params
    res = requests.get(url, **request_params)
    res.raise_for_status()
    return model_from_owl_str(res.text)


def model_from_pc_query(kind, source, target=None, **query_params):
    """Return a BioPAX Model from a Pathway Commons query.

    For more information on these queries, see
    http://www.pathwaycommons.org/pc2/#graph

    Parameters
    ----------
    kind : str
        The kind of graph query to perform. Currently 3 options are
        implemented, 'neighborhood', 'pathsbetween' and 'pathsfromto'.
    source : list[str]
        A single gene name or a list of gene names which are the source set for
        the graph query.
    target : Optional[list[str]]
        A single gene name or a list of gene names which are the target set for
        the graph query. Only needed for 'pathsfromto' queries.
    limit : Optional[int]
        This limits the length of the longest path considered in
        the graph query. Default: 1
    organism : Optional[str]
        The organism used for the query. Default: '9606' corresponding
        to human.
    datasource : Optional[list[str]]
        A list of database sources that the query results should include.
        Example: ['pid', 'panther']. By default, all databases are considered.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model obtained from the results of the Pathway Commons query.
    """
    owl_str = graph_query(kind, source, target=target, **query_params)
    return model_from_owl_str(owl_str)


def model_from_netpath(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a `NetPath <http://netpath.org>`_ entry.

    Parameters
    ----------
    identifier :
        The NetPath identifier for a pathway (e.g., ``22`` for the `leptin
        signaling pathway <http://netpath.org/pathways?path_id=NetPath_22>`_

    Returns
    -------
    :
        A BioPAX model obtained from the NetPath resource.
    """
    url = f"http://netpath.org/data/biopax/NetPath_{identifier}.owl"
    return model_from_owl_url(url)


def model_from_reactome(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a Reactome entry (pathway, event, etc.).

    Parameters
    ----------
    identifier :
        The Reactome identifier for a pathway (e.g., ``177929`` for `Signaling
        by EGFR <https://reactome.org/content/detail/R-HSA-177929>`_)
        or reaction (e.g., ``177946`` for `Pro-EGF is cleaved to form mature
        EGF <https://reactome.org/content/detail/R-HSA-177946>`_). For human
        pathways, the identifier for the BioPAX download is the same as the part
        that comes after ``R-HSA-``. For non-human pathways, this is not so
        clear.

    Returns
    -------
    :
        A BioPAX model obtained from the Reactome resource.
    """
    if identifier.startswith("R-HSA-"):
        # If you give something like R-XXX-YYYYY, just get the YYYYY part back
        # for download.
        identifier = identifier[len("R-HSA-"):]
    url = f"https://reactome.org/ReactomeRESTfulAPI/RESTfulWS/biopaxExporter/" \
          f"Level3/{identifier}"
    return model_from_owl_url(url)


def model_from_humancyc(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a HumanCyc entry.

    Parameters
    ----------
    identifier :
        The HumanCyc identifier for a pathway (e.g., ``PWY66-398`` for
        `TCA cycle
        <https://humancyc.org/HUMAN/NEW-IMAGE?type=PATHWAY&object=PWY66-398>`_)

    Returns
    -------
    :
        A BioPAX model obtained from the HumanCyc pathway.
    """
    return _model_from_xcyc("https://humancyc.org/HUMAN/pathway-biopax",
                            identifier)


def model_from_biocyc(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a `BioCyc <https://biocyc.org>`_ entry.

    BioCyc contains pathways for model eukaryotes and microbes.

    Parameters
    ----------
    identifier :
        The BioCyc identifier for a pathway (e.g., ``P105-PWY`` for
        `TCA cycle IV
        (2-oxoglutarate decarboxylase) <https://biocyc.org/META/NEW-IMAGE?
        type=PATHWAY&object=P105-PWY>`_)

    Returns
    -------
    :
        A BioPAX model obtained from the BioCyc pathway.
    """
    return _model_from_xcyc("https://biocyc.org/META/pathway-biopax",
                            identifier)


def model_from_metacyc(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a `MetaCyc <https://metacyc.org/>`_ entry.

    MetaCyc contains pathways for all organisms

    Parameters
    ----------
    identifier :
        The MetaCyc identifier for a pathway (e.g., ``TCA`` for
        `TCA cycle I (prokaryotic) <https://metacyc.org/META/NEW-IMAGE?type=PATHWAY&object=TCA>`_)

    Returns
    -------
    :
        A BioPAX model obtained from the MetaCyc pathway.
    """
    return _model_from_xcyc("https://metacyc.org/META/pathway-biopax",
                            identifier)


def model_from_ecocyc(identifier: str) -> BioPaxModel:
    """Return a BioPAX model from a `EcoCyc <https://ecocyc.org/>`_ entry.

    EcoCyc contains pathways for Escherichia coli K-12 MG1655.

    Parameters
    ----------
    identifier :
        The EcoCyc identifier for a pathway (e.g., ``TCA`` for
        `TCA cycle I (prokaryotic) <https://ecocyc.org/ECOLI/NEW-IMAGE?type=PATHWAY&object=TCA>`_)

    Returns
    -------
    :
        A BioPAX model obtained from the EcoCyc pathway.
    """
    return _model_from_xcyc("https://ecocyc.org/ECOLI/pathway-biopax",
                            identifier)


def _model_from_xcyc(url: str, identifier: str) -> BioPaxModel:
    """Return a BioPAX model from one of the Cyc databases entry.

    Parameters
    ----------
    url :
        The base url for the XXXCyc BioPAX download endpoint. All of them have
        the form ``https://....../META/pathway-biopax``.
    identifier :
        The site-specific identifier for a pathway

    Returns
    -------
    :
        A BioPAX model obtained from the pathway.
    """
    # Extend URL with arguments
    url = url + f'?type=3&object={identifier}'
    # Not sure if the SSL issue is temporary. Remove verify=False later
    return model_from_owl_url(url, request_params={'verify': False})


def model_to_owl_str(model: BioPaxModel) -> str:
    """Return an OWL string serialized from a BioPaxModel object.

    Parameters
    ----------
    model :
        The BioPaxModel to serialize into an OWL string.

    Returns
    -------
    :
        The OWL string for the model.
    """
    return xml_to_str(model.to_xml())


def model_to_owl_file(model: BioPaxModel,
                      fname: Union[str, pathlib.Path, os.PathLike]):
    """Write an OWL string serialized from a BioPaxModel object into a file.

    Parameters
    ----------
    model :
        The BioPaxModel to serialize into an OWL file.
    fname :
        The path to the target OWL file.
    """
    xml_to_file(model.to_xml(), fname)



__all__ = ['BioPaxModel']

from typing import Any, Mapping, Optional

from tqdm.auto import tqdm

from . import *
from ..xml_util import get_id_or_about, get_tag, has_ns, wrap_xml_elements


class BioPaxModel:
    """BioPAX Model.

    Parameters
    ----------
    objects : dict
        A dict of BioPaxObject instances keyed by their URI string.
    xml_base : str
        The XML base namespace for the content being represented.

    Attributes
    ----------
    objects : dict
        A dict of BioPaxObject instances keyed by their URI string
        that are part of the model.
    xml_base : str
        The XML base namespace for the content being represented.
    """

    def __init__(self, objects, xml_base):
        self.objects = objects
        self.xml_base = xml_base
        self.add_reverse_links()

    @classmethod
    def from_xml(cls, tree) -> "BioPaxModel":
        """Return a BioPAX Model from an OWL/XML element tree.

        Parameters
        ----------
        tree :
            An element tree from which the model is extracted

        Returns
        -------
        :
            A BioPAX Model deserialized from the OWL XML tree.
        """
        objects = {}

        tqdm_kwargs = {'desc': 'Processing OWL elements', "unit_scale": True}
        for element in tqdm(tree, **tqdm_kwargs):
            if not has_ns(element, 'bp'):
                continue
            id = get_id_or_about(element)
            obj_cls = globals()[get_tag(element)]
            obj = obj_cls.from_xml(element)
            objects[id] = obj
            # We now register objects that were recursively
            # extracted but have not been registered yet
            sub_objs = get_sub_objects(obj)
            for sub_obj in sub_objs:
                if sub_obj.uid not in objects:
                    objects[sub_obj.uid] = sub_obj

        for obj_id, obj in objects.items():
            for attr in [a for a in dir(obj) if not a.startswith('_')]:
                # This is to avoid properties
                if attr not in obj.__dict__:
                    continue
                val = getattr(obj, attr)
                resolved_val = resolve_value(objects, val)
                setattr(obj, attr, resolved_val)

        return cls(objects, tree.base)

    def to_xml(self) -> str:
        """Return an OWL string from the content of the model."""
        tqdm_kwargs = {'desc': 'Serializing OWL elements'}
        elements = [obj.to_xml() for obj in tqdm(self.objects.values(),
                                                 **tqdm_kwargs)]
        return wrap_xml_elements(elements, self.xml_base)

    def get_objects_by_type(self, obj_type):
        for obj in self.objects.values():
            if isinstance(obj, obj_type):
                yield obj

    def add_reverse_links(self):
        for uid, obj in self.objects.items():
            for attr in [a for a in dir(obj) if not a.startswith('_')
                         and a not in {'list_types', 'xml_types',
                                       'to_xml', 'from_xml', 'uid'}]:
                val = getattr(obj, attr)
                if isinstance(val, BioPaxObject) or \
                        (isinstance(val, list) and
                         all(isinstance(v, BioPaxObject) for v in val)):
                    if attr in ['left', 'right']:
                        of_attr = '_participant_of'
                    else:
                        of_attr = '_%s_of' % attr
                    vals = val if isinstance(val, list) else [val]
                    for v in vals:
                        if of_attr in dir(v):
                            of_attr_val = getattr(v, of_attr)
                            of_attr_val.add(obj)


def get_sub_objects(obj):
    """Get all the children of an object that were extracted and
    are BioPaxObjects that need to be registered in the model."""
    sub_objs = []
    for attr in [a for a in dir(obj) if not a.startswith('__')]:
        val = getattr(obj, attr)
        if isinstance(val, BioPaxObject):
            sub_objs.append(val)
            sub_objs += get_sub_objects(val)
        elif isinstance(val, list):
            for elem in val:
                if isinstance(elem, BioPaxObject):
                    sub_objs.append(elem)
                    sub_objs += get_sub_objects(elem)
    return sub_objs


def resolve_value(objects, val):
    if isinstance(val, Unresolved):
        if val.obj_id not in objects:
            resolved_val = val.obj_id
        else:
            resolved_val = objects[val.obj_id]
    elif isinstance(val, list):
        resolved_val = [resolve_value(objects, v) for v in val]
    else:
        resolved_val = val
    return resolved_val



__all__ = ['PhysicalEntity', 'SimplePhysicalEntity', 'Protein',
           'SmallMolecule', 'Rna', 'Complex', 'Dna', 'DnaRegion',
           'RnaRegion']

from typing import List, Optional

from .base import Entity, Controller
from .util import EntityFeature, EntityReference


class PhysicalEntity(Entity, Controller):
    """BioPAX PhysicalEntity."""
    list_types = Entity.list_types + \
        ['feature', 'not_feature', 'member_physical_entity']

    def __init__(self,
                 feature: Optional[List[EntityFeature]] = None,
                 not_feature=None,
                 member_physical_entity=None,
                 cellular_location=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.feature = feature
        self.not_feature = not_feature
        self.member_physical_entity = member_physical_entity
        self.cellular_location = cellular_location
        self._component_of = set()
        self._member_physical_entity_of = set()

    @property
    def component_of(self):
        return self._component_of

    @property
    def member_physical_entity_of(self):
        return self._member_physical_entity_of

    def __str__(self):
        name = self.display_name if self.display_name else self.standard_name
        s = '%s(%s)' % (self.__class__.__name__, name)
        return s

    def __repr__(self):
        return str(self)


class SimplePhysicalEntity(PhysicalEntity):
    """BioPAX SimplePhysicalEntity."""
    def __init__(
        self,
        entity_reference: Optional[EntityReference] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.entity_reference = entity_reference


class Protein(SimplePhysicalEntity):
    """BioPAX Protein."""
    pass


class SmallMolecule(SimplePhysicalEntity):
    """BioPAX SmallMolecule."""
    pass


class Rna(SimplePhysicalEntity):
    """BioPAX Rna."""
    pass


class Complex(PhysicalEntity):
    """BioPAX Complex."""
    list_types = PhysicalEntity.list_types + \
        ['component', 'component_stoichiometry']

    def __init__(self,
                 component: Optional[List[PhysicalEntity]] = None,
                 component_stoichiometry: Optional[List] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.component = component
        self.component_stoichiometry = component_stoichiometry


class Dna(SimplePhysicalEntity):
    """BioPAX Dna."""
    pass


class DnaRegion(SimplePhysicalEntity):
    """BioPAX DnaRegion"""
    pass


class RnaRegion(SimplePhysicalEntity):
    """BioPAX RnaRegion"""
    pass
